// mayu_pp.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <boost/filesystem/convenience.hpp>
#include <pstade/oven/stream_lines.hpp>
#include <pstade/oven/dropped.hpp>
#include <str_to_num.hpp>

///////////////////////////////////////////////////////////////////////////////
// Include the lexer stuff
#include <boost/wave/cpplexer/cpp_lex_token.hpp>    // token class
#include <boost/wave/cpplexer/cpp_lex_iterator.hpp> // lexer class

typedef string str_t;

namespace mayu_pp
{
  using namespace boost::wave;

  class advanced_preprocessing_hooks
    :   public boost::wave::context_policies::default_preprocessing_hooks
  {
  public:
    advanced_preprocessing_hooks() {}

    //* no more needed, because editing cpp.re as # be treated as comment. <- This didn't work, because the preprocessor has to recognize # as stringizer or concatenator.
    template <typename ContextT, typename ContainerT>
    bool
      found_unknown_directive(ContextT const& ctx, ContainerT const& line, 
      ContainerT& pending)
    {
      pending = line;
      return true; 
    }
    //*/

    //When the included file's extension isn't the same as the initial top parsing file, output is discarded.
    template <typename ContextT>
    void 
      opened_include_file(ContextT const& ctx, std::string const& relname, 
      std::string const& absname, bool is_system_include) 
    {
      discarding_stack.push(!iequals(top_file_extension, filesystem::extension(relname)));
    }
    template <typename ContextT>
    void
      returning_from_include_file(ContextT const& ctx) 
    {
      discarding_stack.pop();
    }
    bool is_discarding()
    {
      return !discarding_stack.empty() && discarding_stack.top();
    }

    stack<bool> discarding_stack;
    string top_file_extension;
  };

  //Default load_file_to_string just uses ifstream and can't read Unicode files.
  struct load_file_to_string 
  {
    template <typename IterContextT>
    class inner 
    {
    public:
      template <typename PositionT>
      static void init_iterators(IterContextT &iter_ctx, 
        PositionT const &act_pos, language_support language)
      {
        typedef typename IterContextT::iterator_type iterator_type;

        // read in the file
        if (!file::load(iter_ctx.filename.c_str(), iter_ctx.instring)) {
          BOOST_WAVE_THROW_CTX(iter_ctx.ctx, preprocess_exception, 
            bad_include_file, iter_ctx.filename.c_str(), act_pos);
          return;
        }

        iter_ctx.first = iterator_type(
          iter_ctx.instring.begin(), iter_ctx.instring.end(), 
          PositionT(iter_ctx.filename), language);
        iter_ctx.last = iterator_type();
      }

    private:
      str_t instring;
    };
  };
}

int _tmain(int argc, _TCHAR* argv[])
{
  if(argc == 1)
  {
    std::cerr << "An input file must be specified as a parameter." << std::endl;
    return 1;
  }

  boost::wave::util::file_position_type current_position;

  try {
    str_t instring;
    string file_name_a = sconv::t2a(argv[1]);
    wstring file_name = sconv::t2w(argv[1]);

    if(!file::load(file_name, instring)){
      std::cerr << "Could not open input file: " << file_name << std::endl;
      return -2;
    }
    //  This token type is one of the central types used throughout the library, 
    //  because it is a template parameter to some of the public classes and  
    //  instances of this type are returned from the iterators.
    typedef boost::wave::cpplexer::lex_token<> token_type;

    //  The template boost::wave::cpplexer::lex_iterator<> is the lexer type to
    //  to use as the token source for the preprocessing engine. It is 
    //  parametrized with the token type.
    typedef boost::wave::cpplexer::lex_iterator<token_type> lex_iterator_type;

    //  This is the resulting context type to use. The first template parameter
    //  should match the boost::wave::cpplexer::iterator type to be used during construction of the
    //  corresponding context object (see below).
    typedef boost::wave::context<str_t::iterator
                                ,lex_iterator_type
                                ,mayu_pp::load_file_to_string
                                ,mayu_pp::advanced_preprocessing_hooks
    > context_type;

    //  The preprocessor iterator shouldn't be constructed directly. It is 
    //  to be generated through a wave::context<> object. This wave:context<> 
    //  object is to be used additionally to initialize and define different 
    //  parameters of the actual preprocessing (not done here).
    //
    //  The preprocessing of the input stream is done on the fly behind the 
    //  scenes during iteration over the context_type::iterator_type stream.
    context_type ctx (instring.begin(), instring.end(), str_source(file_name_a));
    int support_option_enable = wave::support_option_variadics|wave::support_option_long_long|wave::support_option_preserve_comments|wave::support_option_no_character_validation;
    int support_option_disable = wave::support_option_single_line|wave::support_option_emit_line_directives; //line directives should be emitted?
    ctx.set_language((wave::language_support)((ctx.get_language()|support_option_enable)^support_option_disable));
    ctx.get_hooks().top_file_extension = filesystem::extension(file_name_a);

    //Read mayu_pp.cfg file.
    try
    {
      ifstream f("mayu_pp.cfg");
      BOOST_FOREACH(string s, stream_lines(f))
      {
        trim(s);
        if(s.empty() || s[0] == '#') //# starts a comment line.
          continue;
        if(s[0] != '-')
          throw runtime_error("Config line must start with '-': " + s);
        if(s.size() == 1)
          throw runtime_error("Config line must be followed after '-': " + s);
        switch(s[1])
        {
        case 'S':
          ctx.add_sysinclude_path(s.c_str() + 2);
          break;
        case 'D':
          ctx.add_macro_definition(s.c_str() + 2, false);
          break;
        default:
          throw runtime_error("Config line starting only -S or -D is supported: " + s);
        }
      }
    }
    catch(runtime_error const& e)
    {
      cerr << e.what() << endl;
      return 5;
    }


    //  Get the preprocessor iterators and use them to generate 
    //  the token sequence.
    context_type::iterator_type first = ctx.begin();
    context_type::iterator_type last = ctx.end();


    //  The input stream is preprocessed for you while iterating over the range
    //  [first, last)
    _setmode(_fileno(stdout), _O_BINARY); //needs to prevent \n -> \r\n conversion. Comment tokens have \r\n and output would be \r\r\n.
    while (first != last) {
      current_position = (*first).get_position();
      if(wave::T_ANY == (wave::token_id)*first) //UniversalCharacter \u0000 or \U00000000. To output newlines from macros, I decided use them. T_SPACE2 is used just because it isn't used by cpp_re.re.
        cout << str_to_num<str_t::value_type>(first->get_value()|oven::dropped(2), 0x10);
      else if(!ctx.get_hooks().is_discarding())
        std::cout << (*first).get_value();
      ++first;
    }
    //]
  }
  catch (boost::wave::cpp_exception const& e) {
    // some preprocessing error
    std::cerr 
      << e.file_name() << "(" << e.line_no() << "): "
      << e.description() << std::endl;
    return 2;
  }
  catch (std::exception const& e) {
    // use last recognized token to retrieve the error position
    std::cerr 
      << current_position.get_file() 
      << "(" << current_position.get_line() << "): "
      << "exception caught: " << e.what()
      << std::endl;
    return 3;
  }
  catch (...) {
    // use last recognized token to retrieve the error position
    std::cerr 
      << current_position.get_file() 
      << "(" << current_position.get_line() << "): "
      << "unexpected exception caught." << std::endl;
    return 4;
  }
  return 0;
}

